use strict;
use POSIX qw(strftime);
use DB_File::Lock;
use Mail::Sender;
chdir '/home/kleptog/perl/ddtss' or die;
my $db_hash = "ddtss.db";

my %hash;

# Open DB for write mode
sub DDTSS_Open_Write
{
  my $db = tie %hash,  'DB_File::Lock', $db_hash, O_RDWR, 0666, $DB_HASH, "write";
  if( not defined $db )
  { die "Unable to open DB file for writing ($!)\n" }
  return $db;
}

# Open DB for read mode
sub DDTSS_Open_Read
{
  my $db = tie %hash,  'DB_File::Lock', $db_hash, O_RDONLY, 0666, $DB_HASH, "read";
  if( not defined $db )
  { die "Unable to open DB file for reading ($!)\n" }
  return $db;
}

# close DB, undef caller reference
sub DDTSS_Close
{
  undef $_[0];
  untie %hash;
}

# Get list of untranslated packages
sub DDTSS_Get_Untranslated
{
  my $db = shift;
  my ($key,$value);
  my @todo;
  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # $1 = prio, $2 = package
    next if $key !~ m,^untranslated/(\d+)/(.*),;

    push @todo, [$1,$2];
  }
  return [ map { $_->[1] } sort { $a->[0] <=> $b->[0] } @todo ];
}

# Get list of packages for review
sub DDTSS_Get_ForReview
{
  my $db = shift;
  my $user = shift;
  
  my ($key,$value);
  my @todo;
  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # $1 = package
    next if $key !~ m,^forreview/([^/]+)/owner,;
    my $package = $1;
    my $str = "";
    if( $value eq $user )
    { $str = " (owner)" }
    else
    {
      my $reviewers;
      if( $db->get("forreview/$package/reviewers", $reviewers) == 0 )
      {
        if( scalar( grep { $_ eq $user } split /,/, $reviewers ) )
        { $str = " (reviewed)" }
      }
    }
    push @todo, [$package,$str];
  }
  return [ sort { $a cmp $b } @todo ];
}

# Given a package name, find the key in the database that represents the untranslated description
sub DDTSS_Untranslated_Get_Key
{
  my $db = shift;
  my $package = shift;

  my ($key,$value);
  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    return $key if $key =~ m,^untranslated/\d+/$package,;
  }
  return undef;
}

# Does a sort of soft-lock on a key
sub DDTSS_TryLock
{
  my $db = shift;
  my $key = shift;
  my $user = shift;
  
  my $field;  
  my($lockuser,$locktime);
  if( $db->get( "lock/$key", $field ) )
  { $db->del( "lock/$key" ); $lockuser = ""; $locktime = 0 }
  else
  {
    ($lockuser,$locktime) = split /,/, $field;
  }
  
  if( ($locktime > time() - 900) and ($lockuser ne $user) )
  {
    return 1;
  }
  # May update the time
  $db->put( "lock/$key", $user.",".time() );
  return 0;
}

# Release lock
sub DDTSS_Unlock
{
  my $db = shift;
  my $key = shift;
  my $user = shift;
  
  my $field;
  if( $db->get( "lock/$key", $field ) == 0 )
  {
    my ($lockuser,$locktime) = split /,/, $field;
    if( $lockuser eq $user )
    {
      $db->del( "lock/$key" );
    }
  }
}

# When a new description comes in, we delete the oldest request
sub DDTSS_Kill_Oldest_Request
{
  my $db = shift;
  my $oldest = 2**32;

  my($key,$value);
  
  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # Found a sent request
    if( $key =~ m,^requests/, )
    {
      if( $value < $oldest )
      { $oldest = $value }
    }
  }
  
  $db->del( "requests/$oldest" );
  return;
}

sub DDTSS_Load_Wordlist
{
  my $lang = shift;
  my $fh;
  open $fh, "words-$lang.txt" or return undef;
  
  my $wordlist = {};
  
  while(<$fh>)
  {
    chomp;
    my @f = split /\t/, $_, 2;
    $wordlist->{$f[0]} = $f[1];
  }
  
  return $wordlist;
}

sub DDTSS_Get_Translated
{
  my $db = shift;

  my($key,$value);
  my @trans;  
  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # $1 = state (ok|check), $2 = package, $3 = timestamp
    if( $key =~ m,^logs/(\w+)/([\w-]+)/(\d+), )
    {
      push @trans, [$3, $2, $1, strftime( "%a %b %e %H:%M:%S %Y", localtime $3), $key ];
    }
  }

  @trans = map { [@$_[1,2,3,4]] } sort { $b->[0] <=> $a->[0] } @trans;
  splice @trans, 10;   # Keep only the last 10

#  use Data::Dumper;
#  print Dumper(\@trans);
  return \@trans;  
}

###################################################
## Processing stuff

# Check the number of untranslated messages and if below the minimum, send off a request for a new one
sub process_incoming
{
  my $db = DDTSS_Open_Read();
  
  my $min;
  
  $db->get("config/minuntranslated", $min);
  
  my $queued = 0;
  my ($key, $value);

  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # Found untranslated data
    if( $key =~ m,^untranslated/, )
    { $queued++ }
    # Found a sent request
    if( $key =~ m,^requests/, )
    {
      if( $value > (time() - 1800) )  # Request sent in the last half hour
      {
        $queued++;
      }
    }
  }
  
  DDTSS_Close($db);
  
  # sufficient found
  return if( $queued >= $min );

  $db = DDTSS_Open_Write();
  
  my $sender = setup_email_connection($db);
  my ($lang,$charset);
  $db->get("config/lang", $lang );
  $db->get("config/charset", $charset );
  
  # Send mail
  eval {
    $sender->MailMsg({subject => "GET 1 $lang.$charset", msg => "NT"});
    -t STDERR and print STDERR "Sent request for new translation\n";
  };
  if( not $@ )
  {
    # Record time sent
    $db->put("requests/".time(), time() );
  }  
  else
  {
    -t STDERR and print STDERR "Send email failed ($@)\n";
  }
  
  DDTSS_Close($db);
}

# Go through the reviewed translations and send them if they've got enough reviews
sub process_reviewed
{
  my $db = DDTSS_Open_Read();

  my ($key, $value);
  my (@todo,@pending);

  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # $1 = package
    next if $key !~ m,^forreview/([-\w]+)/owner,;

    push @pending, $1;;
  }

  my $numreviewers;
  $db->get("config/numreviewers", $numreviewers );
  
  if( $numreviewers > 0 )
  {
    # count reivewers for each package
    foreach my $package (@pending)
    {
      my $reviewers;
      if( $db->get("forreview/$package/reviewers",$reviewers) == 0 )
      {
        my @a;
        if( scalar( @a=split /,/,$reviewers ) >= $numreviewers )
        {
          push @todo, $package;
        }
      }
    }
  }
  else
  {
    @todo = @pending;
  }
  
  DDTSS_Close($db);
  
  # Return if nothing to do
  return if scalar(@todo) == 0;

  $db = DDTSS_Open_Write();
  
  my $sender = setup_email_connection($db);
  
  my($lang,$charset);
  
  $db->get("config/lang", $lang );
  $db->get("config/charset", $charset );
  
  for my $package (@todo)
  {
    # Get details of translation
    my( $short,$long,$data,$owner,$reviewers );
    $db->get("forreview/$package/short", $short );
    $db->get("forreview/$package/long", $long );
    $db->get("forreview/$package/data", $data );
    $db->get("forreview/$package/owner", $owner );
    $db->get("forreview/$package/reviewers", $reviewers );
    
    $reviewers ||= '';
  
    # Substitute translated description in
    unless( $data =~ s/^Description-[\w.-]+:(( .*)?\n)+#/Description-$lang.$charset: $short\n$long\n#/m )
    {
      -t STDERR and print STDERR "Substitution failed for $package, not sending\n";
      next;
    }

    # Add our data
    $data = "# Creator: $owner\n# Reviewers: $reviewers\n".$data;
    eval {
      $sender->OpenMultipart();
      $sender->Body({msg =>"Message from a DDTSS\n"});
      $sender->Part({description => "Translation of $package, language $lang",
                     ctype => 'application/debian-dt',
                     encoding => 'Base64',
                     disposition => 'attachment; filename="new.ddtp"',
                     msg => $data });
      $sender->Close;
    };
    if( not $@ )
    {
      -t STDERR and print STDERR "Sent updated description for $package\n";
      # If worked, store result
      $db->put("done/$package/data", $data);
      $db->del("forreview/$package/short");
      $db->del("forreview/$package/owner");
      $db->del("forreview/$package/long");
      $db->del("forreview/$package/data");
      $db->del("forreview/$package/reviewers");
    }                                                             
    else
    {
      -t STDERR and print STDERR "Send email failed ($@)\n";
    }
    
#    print "===\n$data\n===\n";
  }
  
  DDTSS_Close($db);
}

# Process stuff in the todo area
sub process_todo
{
  my $db = DDTSS_Open_Read();

  my ($key, $value);
  my @todo;

  for (my $status = $db->seq($key, $value, R_FIRST) ;
       $status == 0 ;
       $status = $db->seq($key, $value, R_NEXT) )
  {
    # $1 = type, $2 = timestamp
    next if $key !~ m,^todo/([-\w]+)/\d+\.(\d+),;

    push @todo, [$1,$key,$2];
    
    -t STDERR and print STDERR "TODO: $key\n";
  }

  DDTSS_Close($db);

  return if scalar(@todo) == 0;

  $db = DDTSS_Open_Write();
    
  foreach my $todo (@todo)
  {
    my($type,$key,$ts) = @$todo;
    
    if( $type eq "in-log" )  # Incoming log
    { process_log($db,$key,$ts) }
    elsif( $type eq "in-data" )  # Incoming data
    { process_data($db,$key,$ts) }
    else
    { -t STDERR and print STDERR "Warning: Will not process $key\n"; }
  }

  DDTSS_Close($db);
}

# Process a log file sent by the DDTS
sub process_log
{
  my $db = shift;
  my $key = shift;
  my $data;
  return if $db->get($key,$data);  # Someone maybe did it already
  my $timestamp = shift;

  if( $data !~ /start:/ )
  {
    -t STDERR and print STDERR "Not a translation log\n";
    $db->del($key);
    return;
  }
  my ($descr) = ($data =~ /^Description: (.*)$/m);
  
  if( not defined $descr )
  {
    -t STDERR and print STDERR "Cannot find description in log\n";
    return;
  }

  my $package;
  if( $db->get("descrmatch/$descr",$package) )
  {
    -t STDERR and print STDERR "Cannot find package matching description '$descr'\n";
    return;
  }
  
  my $state;
  if( $data =~ /^add the translation in the db$/m )
  { $state = "ok" }
  else
  { $state = "check" };
  
  -t STDERR and print STDERR "Moving '$key' to 'logs/$state/$package/$timestamp\n";
  
  $db->put("logs/$state/$package/$timestamp", $data);
  $db->del($key);
}

# Process new translation sent by DDTS
sub process_data
{
  my $db = shift;
  my $key = shift;
  my $data;
  return if $db->get($key,$data);  # Someone maybe did it already
  my $timestamp = shift;

  my ($package) = ($data =~ /^# Package[^:]*: (.*)$/m);
  
  if( not defined $package )
  {
    -t STDERR and print STDERR "Cannot find package name in data\n";
    return;
  }

  my ($prio) = ($data =~ /^# Prioritize: (\d+)$/m);
  
  if( not defined $prio ) { $prio = 30 }

  my ($descr) = ($data =~ /^Description: (.*)$/m);

  if( not defined $descr )
  {
    -t STDERR and print STDERR "Cannot find description in data\n";
    return;
  }

  -t STDERR and print STDERR  "Moving '$key' to 'untranslated/$prio/$package'\n";
  
  DDTSS_Kill_Oldest_Request($db);
  
  $db->put("untranslated/$prio/$package", $data);
  $db->put("descrmatch/$descr",$package);
  $db->del($key);
}

# Setup Mail::Sender object
sub setup_email_connection
{
  my $db = shift;
  my ($from,$to,$mailserver,$emaildebug);
  $db->get("config/clientemail",$from);
  $db->get("config/serveremail",$to);
  $db->get("config/mailserver",$mailserver);
  $db->get("config/debugemail",$emaildebug);
  
  if( not defined $mailserver )
  { $mailserver = "127.0.0.1" }
  
  my $sender = new Mail::Sender { smtp => $mailserver, 
                                 from => $from, 
                                 to => $to, 
                                 subject => 'nothing', 
                                 on_errors => 'die',
                                 (defined $emaildebug ? (bcc => $emaildebug) : ()) };
                               
  return $sender;
}


1;